#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.converter module

This module provides all default schema fields converters.
"""

import datetime
import decimal
from cgi import FieldStorage

import six
from zope.i18n.format import DateTimeParseError, NumberParseError
from zope.interface import alsoProvides, implementer
try:
    from zope.publisher.browser import FileUpload
except ImportError:
    FileUpload = None
from zope.schema import ValidationError
from zope.schema.interfaces import IBool, IBytes, ICollection, IDate, IDatetime, IDecimal, IDict, \
    IField, IFloat, IFromUnicode, IInt, ISequence, ITime, ITimedelta

from pyams_form.interfaces import IDataConverter
from pyams_form.interfaces.form import IFormAware
from pyams_form.interfaces.widget import IFieldWidget, IMultiWidget, ISequenceWidget, \
    ISingleCheckBoxWidget, ITextLinesWidget, IWidget, IFileWidget
from pyams_form.util import to_bytes, to_unicode
from pyams_utils.adapter import adapter_config
from pyams_utils.interfaces.form import NOT_CHANGED, NO_VALUE

__docformat__ = 'restructuredtext'

from pyams_form import _  # pylint: disable=ungrouped-imports


@implementer(IDataConverter)
class BaseDataConverter:
    """A base implementation of the data converter."""

    _strip_value = True  # Remove spaces at start and end of text line

    def __init__(self, field, widget):
        self.field = field
        self.widget = widget

    def _get_converter(self, field):
        # We rely on the default registered widget, this is probably a
        # restriction for custom widgets. If so use your own MultiWidget and
        # register your own converter which will get the right widget for the
        # used value_type.
        registry = self.widget.request.registry
        widget = registry.getMultiAdapter((field, self.widget.request),
                                          IFieldWidget)
        if IFormAware.providedBy(self.widget):
            # form property required by object widget
            widget.form = self.widget.form
            alsoProvides(widget, IFormAware)
        converter = registry.getMultiAdapter((field, widget), IDataConverter)
        return converter

    def to_widget_value(self, value):
        """See interfaces.IDataConverter"""
        if value is self.field.missing_value:
            return ''
        return to_unicode(value)

    def to_field_value(self, value):
        """See interfaces.IDataConverter"""
        if self._strip_value and isinstance(value, six.string_types):
            value = value.strip()

        if value == '':
            return self.field.missing_value

        # this is going to burp with `Object is of wrong type.`
        # if a non-unicode value comes in from the request
        return self.field.fromUnicode(value)

    def __repr__(self):
        return '<%s converts from %s to %s>' % (
            self.__class__.__name__,
            self.field.__class__.__name__,
            self.widget.__class__.__name__)


@adapter_config(required=(IField, IWidget), provides=IDataConverter)
class FieldDataConverter(BaseDataConverter):
    """A data converter using the field's ``fromUnicode()`` method."""

    def __init__(self, field, widget):
        super().__init__(field, widget)
        if not IFromUnicode.providedBy(field):
            field_name = ''
            if field.__name__:
                field_name = '``%s`` ' % field.__name__
            raise TypeError('Field %s of type ``%s`` must provide ``IFromUnicode``.' %
                            (field_name, type(field).__name__))


@adapter_config(required=IFieldWidget, provides=IDataConverter)
def FieldWidgetDataConverter(widget):  # pylint: disable=invalid-name
    """Provide a data converter based on a field widget."""
    return widget.request.registry.queryMultiAdapter((widget.field, widget), IDataConverter)


class FormatterValidationError(ValidationError):
    """Formatter validation error"""

    message = None

    def __init__(self, message, value):
        ValidationError.__init__(self, message, value)
        self.message = message

    def doc(self):
        return self.message


class NumberDataConverter(BaseDataConverter):
    """A general data converter for numbers."""

    type = None
    error_message = None

    def __init__(self, field, widget):
        super().__init__(field, widget)
        locale = self.widget.request.locale
        self.formatter = locale.numbers.getFormatter('decimal')
        self.formatter.type = self.type

    def to_widget_value(self, value):
        """See interfaces.IDataConverter"""
        if value == self.field.missing_value:
            return ''
        return self.formatter.format(value)

    def to_field_value(self, value):
        """See interfaces.IDataConverter"""
        if value == '':
            return self.field.missing_value
        try:
            return self.formatter.parse(value)
        except NumberParseError as exc:
            raise FormatterValidationError(self.error_message, value) from exc


@adapter_config(required=(IInt, IWidget), provides=IDataConverter)
class IntegerDataConverter(NumberDataConverter):
    """A data converter for integers."""

    type = int
    error_message = _('The entered value is not a valid integer literal.')


@adapter_config(required=(IFloat, IWidget), provides=IDataConverter)
class FloatDataConverter(NumberDataConverter):
    """A data converter for integers."""

    type = float
    error_message = _('The entered value is not a valid decimal literal.')


@adapter_config(required=(IDecimal, IWidget), provides=IDataConverter)
class DecimalDataConverter(NumberDataConverter):
    """A data converter for integers."""

    type = decimal.Decimal
    error_message = _('The entered value is not a valid decimal literal.')


class CalendarDataConverter(BaseDataConverter):
    """A special data converter for calendar-related values."""

    type = None
    length = 'short'

    def __init__(self, field, widget):
        super().__init__(field, widget)
        locale = self.widget.request.locale
        self.formatter = locale.dates.getFormatter(self.type, self.length)

    def to_widget_value(self, value):
        """See interfaces.IDataConverter"""
        if value is self.field.missing_value:
            return ''
        return self.formatter.format(value)

    def to_field_value(self, value):
        """See interfaces.IDataConverter"""
        if value == '':
            return self.field.missing_value
        try:
            return self.formatter.parse(value)
        except DateTimeParseError as err:
            raise FormatterValidationError(err.args[0], value) from err


@adapter_config(required=(IDate, IWidget), provides=IDataConverter)
class DateDataConverter(CalendarDataConverter):
    """A special data converter for dates."""

    type = 'date'


@adapter_config(required=(ITime, IWidget), provides=IDataConverter)
class TimeDataConverter(CalendarDataConverter):
    """A special data converter for times."""

    type = 'time'


@adapter_config(required=(IDatetime, IWidget), provides=IDataConverter)
class DatetimeDataConverter(CalendarDataConverter):
    """A special data converter for datetimes."""

    type = 'dateTime'


@adapter_config(required=(ITimedelta, IWidget), provides=IDataConverter)
class TimedeltaDataConverter(FieldDataConverter):
    """A special data converter for timedeltas."""

    def __init__(self, field, widget):
        # pylint: disable=super-init-not-called
        self.field = field
        self.widget = widget

    def to_field_value(self, value):
        """See interfaces.IDataConverter"""
        if value == '':
            return self.field.missing_value
        try:
            days_string, crap, time_string = value.split(' ')  # pylint: disable=unused-variable
        except ValueError:
            time_string = value
            days = 0
        else:
            days = int(days_string)
        seconds = [int(part) * 60 ** (2 - n)
                   for n, part in enumerate(time_string.split(':'))]
        return datetime.timedelta(days, sum(seconds))


@adapter_config(required=(IBytes, IFileWidget), provides=IDataConverter)
class FileUploadDataConverter(BaseDataConverter):
    """A special data converter for bytes, supporting also FileUpload.

    Since IBytes represents a file upload too, this converter can handle
    zope.publisher.browser.FileUpload object as given value.
    """

    def to_widget_value(self, value):
        """See interfaces.IDataConverter"""
        return None

    def to_field_value(self, value):
        """See interfaces.IDataConverter"""
        if value is None or value == '':
            # When no new file is uploaded, send a signal that we do not want
            # to do anything special.
            return NOT_CHANGED

        # By default a IBytes field is used for get a file upload widget.
        # But interfaces extending IBytes do not use file upload widgets.
        # Any way if we get a FieldStorage of FileUpload object, we'll
        # convert it.
        # We also store the additional FieldStorage/FileUpload values on the widget
        # before we loose them.
        if isinstance(value, (FieldStorage, FileUpload)):
            self.widget.headers = value.headers
            self.widget.filename = value.filename
            try:
                if isinstance(value, FieldStorage):
                    if value.fp is None:
                        seek = value.file.seek
                        read = value.file.read
                    else:
                        seek = value.fp.seek
                        read = value.fp.read
                else:
                    seek = value.seek
                    read = value.read
            except AttributeError as e:  # pylint: disable=invalid-name
                raise ValueError(_('Bytes data are not a file object')) from e
            else:
                seek(0)
                data = read()
                if data or getattr(value, 'filename', ''):
                    return data
                return self.field.missing_value
        else:
            return to_bytes(value)


@adapter_config(required=(IField, ISequenceWidget), provides=IDataConverter)
class SequenceDataConverter(BaseDataConverter):
    """Basic data converter for ISequenceWidget."""

    def to_widget_value(self, value):
        """Convert from Python bool to HTML representation."""
        # if the value is the missing value, then an empty list is produced.
        if value is self.field.missing_value:
            return []
        # Look up the term in the terms
        terms = self.widget.update_terms()
        try:
            return [terms.getTerm(value).token]
        except LookupError:
            # Swallow lookup errors, in case the options changed.
            return []

    def to_field_value(self, value):
        """See interfaces.IDataConverter"""
        widget = self.widget
        if (len(value) == 0) or (value[0] == widget.no_value_token):
            return self.field.missing_value
        widget.update_terms()
        return widget.terms.getValue(value[0])


@adapter_config(required=(ICollection, ISequenceWidget), provides=IDataConverter)
class CollectionSequenceDataConverter(BaseDataConverter):
    """A special converter between collections and sequence widgets."""

    def to_widget_value(self, value):
        """Convert from Python bool to HTML representation."""
        if value is self.field.missing_value:
            return []
        widget = self.widget
        if widget.terms is None:
            widget.update_terms()
        values = []
        for entry in value:
            try:
                values.append(widget.terms.getTerm(entry).token)
            except LookupError:
                # Swallow lookup errors, in case the options changed.
                pass
        return values

    def to_field_value(self, value):
        """See interfaces.IDataConverter"""
        if value is NO_VALUE:
            value = []
        widget = self.widget
        if widget.terms is None:
            widget.update_terms()
        collection_type = self.field._type  # pylint: disable=protected-access
        if isinstance(collection_type, tuple):
            collection_type = collection_type[-1]
        return collection_type([widget.terms.getValue(token) for token in value])


@adapter_config(required=(ISequence, ITextLinesWidget), provides=IDataConverter)
class TextLinesConverter(BaseDataConverter):
    """Data converter for ITextLinesWidget."""

    def to_widget_value(self, value):
        """Convert from text lines to HTML representation."""
        # if the value is the missing value, then an empty list is produced.
        if value is self.field.missing_value:
            return ''
        return '\n'.join(to_unicode(v) for v in value)

    def to_field_value(self, value):
        """See interfaces.IDataConverter"""
        collection_type = self.field._type  # pylint: disable=protected-access
        if isinstance(collection_type, tuple):
            collection_type = collection_type[-1]
        if len(value) == 0:
            return self.field.missing_value
        value_type = self.field.value_type._type  # pylint: disable=protected-access
        if isinstance(value_type, tuple):
            value_type = value_type[0]
        # having a blank line at the end matters, one might want to have a blank
        # entry at the end, resp. do not eat it once we have one
        # splitlines ate that, so need to use split now
        value = value.replace('\r\n', '\n')
        items = []
        for val in value.split('\n'):
            try:
                items.append(value_type(val))
            except ValueError as err:
                raise FormatterValidationError(str(err), val) from err
        return collection_type(items)


@adapter_config(required=(ISequence, IMultiWidget), provides=IDataConverter)
class MultiConverter(BaseDataConverter):
    """Data converter for IMultiWidget."""

    def to_widget_value(self, value):
        """Just dispatch it."""
        if value is self.field.missing_value:
            return []
        converter = self._get_converter(self.field.value_type)

        # we always return a list of values for the widget
        return [converter.to_widget_value(v) for v in value]

    def to_field_value(self, value):
        """Just dispatch it."""
        if len(value) == 0:
            return self.field.missing_value

        converter = self._get_converter(self.field.value_type)
        values = [converter.to_field_value(v) for v in value]

        # convert the field values to a tuple or list
        collection_type = self.field._type  # pylint: disable=protected-access
        return collection_type(values)


@adapter_config(required=(IDict, IMultiWidget), provides=IDataConverter)
class DictMultiConverter(BaseDataConverter):
    """Data converter for IMultiWidget."""

    def to_widget_value(self, value):
        """Just dispatch it."""
        if value is self.field.missing_value:
            return []

        key_converter = self._get_converter(self.field.key_type)
        converter = self._get_converter(self.field.value_type)
        # we always return a list of values for the widget
        return [(key_converter.to_widget_value(k), converter.to_widget_value(v))
                for k, v in value.items()]

    def to_field_value(self, value):
        """Just dispatch it."""
        if len(value) == 0:
            return self.field.missing_value

        key_converter = self._get_converter(self.field.key_type)
        converter = self._get_converter(self.field.value_type)
        return {key_converter.to_field_value(k): converter.to_field_value(v)
                for k, v in value}


@adapter_config(required=(IBool, ISingleCheckBoxWidget), provides=IDataConverter)
class BoolSingleCheckboxDataConverter(BaseDataConverter):
    "A special converter between boolean fields and single checkbox widgets."

    def to_widget_value(self, value):
        """Convert from Python bool to HTML representation."""
        if value:
            return ['selected']
        return []

    def to_field_value(self, value):
        """See interfaces.IDataConverter"""
        if value and value[0] == 'selected':
            return True
        return False



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.subform module

This module provides subforms management.
"""

from zope.interface import implementer
from zope.lifecycleevent import ObjectModifiedEvent

from pyams_form.button import handler
from pyams_form.form import BaseForm, EditForm, apply_changes
from pyams_form.interfaces import DISPLAY_MODE
from pyams_form.interfaces.button import IActionHandler
from pyams_form.interfaces.form import IHandlerForm, IInnerForm, ISubForm


__docformat__ = 'restructuredtext'

from pyams_form import _  # pylint: disable=ungrouped-imports


@implementer(ISubForm, IInnerForm)
class BaseInnerForm(BaseForm):
    """Base inner subform"""

    def __init__(self, context, request, parent_form):
        super().__init__(context, request)
        self.parent_form = self.__parent__ = parent_form


class InnerAddForm(BaseInnerForm):
    """Inner subform into an add form"""

    ignore_context = True


class InnerEditForm(BaseInnerForm):
    """Inner edit subform into a main edit form"""


class InnerDisplayForm(BaseInnerForm):
    """Inner display form"""

    _mode = DISPLAY_MODE


@implementer(ISubForm, IHandlerForm)
class EditSubForm(BaseForm):
    """Edit sub-form"""

    form_errors_message = _('There were some errors.')
    success_message = _('Data successfully updated.')
    no_changes_message = _('No changes were applied.')

    def __init__(self, context, request, parent_form):
        super().__init__(context, request)
        self.parent_form = self.__parent__ = parent_form

    @handler(EditForm.buttons['apply'])
    def handle_apply(self, action):  # pylint: disable=unused-argument
        """Handler for apply button"""
        data, errors = self.widgets.extract()
        if errors:
            self.status = self.form_errors_message
            return
        content = self.get_content()
        changed = apply_changes(self, content, data)
        if changed:
            registry = self.request.registry
            registry.notify(ObjectModifiedEvent(content))
            self.status = self.success_message
        else:
            self.status = self.no_changes_message

    def update(self):
        super().update()
        registry = self.request.registry
        for action in self.parent_form.actions.executed_actions:
            adapter = registry.queryMultiAdapter((self, self.request, self.get_content(), action),
                                                 IActionHandler)
            if adapter:
                adapter()



#
# Copyright (c) 2007 Zope Foundation and Contributors.
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser.select module

Select widget implementation.
"""

from zope.interface import Interface, implementer_only
from zope.schema.interfaces import IChoice, ITitledTokenizedTerm, IUnorderedCollection

from pyams_form.browser.widget import HTMLSelectWidget, add_field_class
from pyams_form.interfaces.widget import IFieldWidget, ISelectWidget
from pyams_form.widget import FieldWidget, SequenceWidget
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config


__docformat__ = "reStructuredText"

from pyams_form import _  # pylint: disable=ungrouped-imports


@implementer_only(ISelectWidget)
class SelectWidget(HTMLSelectWidget, SequenceWidget):
    """Select widget implementation."""

    klass = 'select-widget'
    css = 'select'
    prompt = False

    no_value_message = _('No value')
    prompt_message = _('Select a value...')

    # Internal attributes
    _adapter_value_attributes = SequenceWidget._adapter_value_attributes + \
        ('no_value_message', 'prompt_message', 'prompt')

    def is_selected(self, term):
        """Check for term selection"""
        return term.token in self.value

    def update(self):
        """See pyams_form.interfaces.widget.IWidget."""
        super().update()
        add_field_class(self)

    @property
    def items(self):
        """Items list getter"""
        if self.terms is None:  # update() has not been called yet
            return ()
        items = []
        if (not self.required or self.prompt) and self.multiple is None:
            if self.prompt:
                message = self.prompt_message
            else:
                message = self.no_value_message
            items.append({
                'id': self.id + '-novalue',
                'value': self.no_value_token,
                'content': message,
                'selected': self.value in ((), [])
            })

        ignored = set(self.value)

        def add_item(idx, term, prefix=''):
            selected = self.is_selected(term)
            if selected and term.token in ignored:
                ignored.remove(term.token)
            item_id = '%s-%s%i' % (self.id, prefix, idx)
            content = term.token
            if ITitledTokenizedTerm.providedBy(term):
                content = self.request.localizer.translate(term.title)
            items.append({
                'id': item_id,
                'value': term.token,
                'content': content,
                'selected': selected
            })

        for idx, term in enumerate(self.terms):
            add_item(idx, term)

        if ignored:
            # some values are not displayed, probably they went away from the vocabulary
            for idx, token in enumerate(sorted(ignored)):
                try:
                    term = self.terms.getTermByToken(token)
                except LookupError:
                    # just in case the term really went away
                    continue

                add_item(idx, term, prefix='missing-')
        return items

    def json_data(self):
        data = super().json_data()
        data['type'] = 'select'
        data['options'] = self.items
        return data


@adapter_config(required=(IChoice, IFormLayer), provided=IFieldWidget)
def ChoiceWidgetDispatcher(field, request):  # pylint: disable=invalid-name
    """Dispatch widget for IChoice based also on its source."""
    return request.registry.getMultiAdapter((field, field.vocabulary, request),
                                            IFieldWidget)


@adapter_config(required=(IChoice, Interface, IFormLayer), provided=IFieldWidget)
def SelectFieldWidget(field, source, request=None):  # pylint: disable=invalid-name
    """IFieldWidget factory for SelectWidget."""
    if request is None:
        real_request = source
    else:
        real_request = request
    return FieldWidget(field, SelectWidget(real_request))


@adapter_config(required=(IUnorderedCollection, IFormLayer), provided=IFieldWidget)
def CollectionSelectFieldWidget(field, request):  # pylint: disable=invalid-name
    """IFieldWidget factory for SelectWidget."""
    widget = request.registry.getMultiAdapter((field, field.value_type, request), IFieldWidget)
    widget.size = 5
    widget.multiple = 'multiple'
    return widget


@adapter_config(required=(IUnorderedCollection, IChoice, IFormLayer), provided=IFieldWidget)
def CollectionChoiceSelectFieldWidget(field, value_type, request):
    # pylint: disable=invalid-name,unused-argument
    """IFieldWidget factory for SelectWidget."""
    return SelectFieldWidget(field, None, request)



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.value module

Simple value adapters.
"""

from zope.component import adapter
from zope.interface import implementer

from pyams_form.interfaces import IValue
from pyams_form.util import get_specification


__docformat__ = 'restructuredtext'


@implementer(IValue)
class StaticValue:
    """Static value adapter."""

    def __init__(self, value):
        self.value = value

    def get(self):
        """Get static value"""
        return self.value

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.value)


@implementer(IValue)
class ComputedValue:
    """Computed value adapter."""

    def __init__(self, func):
        self.func = func

    def get(self):
        """Get computed value"""
        return self.func(self)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.get())


class ValueFactory:
    """Computed value factory."""

    def __init__(self, value, value_class, discriminators):
        self.value = value
        self.value_class = value_class
        self.discriminators = discriminators

    def __call__(self, *args):
        value_class = self.value_class(self.value)
        for name, value in zip(self.discriminators, args):
            setattr(value_class, name, value)
        return value_class


class ValueCreator:
    """Base class for value creator"""

    value_class = StaticValue

    def __init__(self, discriminators):
        self.discriminators = discriminators

    def __call__(self, value, **kws):
        # Step 1: Check that the keyword argument names match the
        #         discriminators
        if set(kws).difference(set(self.discriminators)):
            raise ValueError('One or more keyword arguments did not match the '
                             'discriminators.')
        # Step 2: Create an attribute value factory
        factory = ValueFactory(value, self.value_class, self.discriminators)
        # Step 3: Build the adaptation signature
        signature = []
        for disc in self.discriminators:
            spec = get_specification(kws.get(disc))
            signature.append(spec)
        # Step 4: Assert the adaptation signature onto the factory
        adapter(*signature)(factory)
        implementer(IValue)(factory)
        return factory


class StaticValueCreator(ValueCreator):
    """Creates static value."""

    value_class = StaticValue


class ComputedValueCreator(ValueCreator):
    """Creates computed value."""

    value_class = ComputedValue



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.template module

Widgets templates management module.
"""

import inspect
import os

import venusian
from pyramid.config import ConfigurationError
from pyramid_chameleon.interfaces import IChameleonTranslate
from zope.interface import Interface, directlyProvides

from pyams_form.interfaces import INPUT_MODE, IWidgetLayoutTemplate
from pyams_layer.interfaces import IFormLayer
from pyams_template.interfaces import IPageTemplate
from pyams_template.template import TemplateFactory, ViewTemplate
from pyams_utils.registry import get_pyramid_registry, query_utility


__docformat__ = 'restructuredtext'


class WidgetTemplateFactory(TemplateFactory):
    """Widget template factory"""

    def __call__(self, context, request, form, field, widget):
        # pylint: disable=too-many-arguments
        return self.template


class WidgetTemplate(ViewTemplate):
    """Widget template class"""

    def __init__(self, provides):
        super().__init__(provides)

    def __call__(self, instance, *args, **keywords):
        request = instance.request
        registry = request.registry
        template = registry.getMultiAdapter(
            (request.context, request, instance.form, instance.field, instance),
            self.provides, name=instance.mode)

        keywords.update({
            'context': instance.context,
            'request': instance.request,
            'view': instance,
            'translate': query_utility(IChameleonTranslate)
        })
        return template(*args, **keywords)


class GetWidgetTemplate(WidgetTemplate):
    """Get widget template"""

    def __init__(self):
        super().__init__(IPageTemplate)


get_widget_template = GetWidgetTemplate  # pylint: disable=invalid-name


class GetWidgetLayout(WidgetTemplate):
    """Layout template getter class"""

    def __init__(self):
        super().__init__(IWidgetLayoutTemplate)


get_widget_layout = GetWidgetLayout  # pylint: disable=invalid-name


def register_widget_template(template, widget, settings, provides, registry=None):
    """Register new widget template"""
    if not os.path.isfile(template):
        raise ConfigurationError("No such file", template)

    content_type = settings.get('content_type', 'text/html')
    macro = settings.get('macro')
    factory = WidgetTemplateFactory(template, content_type, macro)
    provides = settings.get('provides', provides)
    directlyProvides(factory, provides)

    # check context
    required = (
        settings.get('context', Interface),
        settings.get('layer', IFormLayer),
        settings.get('form', None),
        settings.get('field', None),
        widget
    )
    # update registry
    if registry is None:
        registry = settings.get('registry')
        if registry is None:
            registry = get_pyramid_registry()
    registry.registerAdapter(factory, required, provides, settings.get('mode', INPUT_MODE))


class base_widget_template_config:  # pylint: disable=invalid-name
    """Class decorator used to declare a widget template"""

    venusian = venusian  # for testing injection
    interface = IPageTemplate  # template interface

    def __init__(self, **settings):
        if 'for_' in settings:
            if settings.get('context') is None:
                settings['context'] = settings['for_']
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        settings = self.__dict__.copy()

        def callback(context, name, widget):  # pylint: disable=unused-argument
            template = os.path.join(os.path.dirname(inspect.getfile(inspect.getmodule(widget))),
                                    settings.get('template'))
            if not os.path.isfile(template):
                raise ConfigurationError("No such file", template)
            # check registry
            registry = settings.get('registry')
            if registry is None:
                config = context.config.with_package(info.module)  # pylint: disable=no-member
                registry = config.registry
            register_widget_template(template, widget, settings, self.interface, registry)

        info = self.venusian.attach(wrapped, callback, category='pyramid')
        if info.scope == 'class':  # pylint: disable=no-member
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if settings.get('attr') is None:
                settings['attr'] = wrapped.__name__

        settings['_info'] = info.codeinfo  # pylint: disable=no-member
        return wrapped


class widget_template_config(base_widget_template_config):  # pylint: disable=invalid-name
    """Class decorator used to declare a widget template"""


def override_widget_template(widget, **settings):
    """Override template for a given widget

    This function can be used to override a widget template without using ZCML.
    Settings are the same as for @widget_template_config decorator.
    """
    template = settings.get('template', '')
    if not template:
        raise ConfigurationError("No template specified")
    if not template.startswith('/'):
        stack = inspect.stack()[1]
        template = os.path.join(os.path.dirname(inspect.getfile(inspect.getmodule(stack[0]))),
                                template)
    register_widget_template(template, widget, settings, IPageTemplate)


class widget_layout_config(base_widget_template_config):  # pylint: disable=invalid-name
    """Class decorator used to declare a layout template"""

    interface = IWidgetLayoutTemplate  # template interface


def override_widget_layout(widget, **settings):
    """Override layout template for a given class

    This function can be used to override a class layout template without using ZCML.
    Settings are the same as for @layout_config decorator.
    """
    template = settings.get('template', '')
    if not template:
        raise ConfigurationError("No template specified")
    if not template.startswith('/'):
        stack = inspect.stack()[1]
        template = os.path.join(os.path.dirname(inspect.getfile(inspect.getmodule(stack[0]))),
                                template)
    register_widget_template(template, widget, settings, IWidgetLayoutTemplate)



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser.checkbox module

Checkbox widget implementation.
"""

from zope.interface import implementer_only
from zope.schema.interfaces import IBool, ITitledTokenizedTerm
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary

from pyams_form.browser.widget import HTMLInputWidget, add_field_class
from pyams_form.interfaces.widget import ICheckBoxWidget, IFieldWidget, ISingleCheckBoxWidget
from pyams_form.term import Terms
from pyams_form.util import to_unicode
from pyams_form.widget import FieldWidget, SequenceWidget
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config


__docformat__ = 'restructuredtext'


@implementer_only(ICheckBoxWidget)
class CheckBoxWidget(HTMLInputWidget, SequenceWidget):
    """Input type checkbox widget implementation."""

    klass = 'checkbox-widget'
    css = 'checkbox'

    def is_checked(self, term):
        """Check if given term is selected"""
        return term.token in self.value

    @property
    def items(self):
        """Items list getter"""
        if self.terms is None:
            return ()
        items = []
        for count, term in enumerate(self.terms):
            checked = self.is_checked(term)
            item_id = '%s-%i' % (self.id, count)
            if ITitledTokenizedTerm.providedBy(term):
                label = self.request.localizer.translate(term.title)
            else:
                label = to_unicode(term.value)
            items.append({
                'id': item_id,
                'name': self.name,
                'value': term.token,
                'label': label,
                'checked': checked
            })
        return items

    def update(self):
        """See pyams_form.interfaces.widget.IWidget."""
        super().update()
        add_field_class(self)

    def json_data(self):
        """Get widget data in JSON format"""
        data = super().json_data()
        data['type'] = 'check'
        data['options'] = list(self.items)
        return data


def CheckBoxFieldWidget(field, request):  # pylint: disable=invalid-name
    """IFieldWidget factory for CheckBoxWidget."""
    return FieldWidget(field, CheckBoxWidget(request))


@implementer_only(ISingleCheckBoxWidget)
class SingleCheckBoxWidget(CheckBoxWidget):
    """Single input type checkbox widget implementation."""

    klass = 'single-checkbox-widget'

    def update_terms(self):
        """Update terms"""
        if self.terms is None:
            self.terms = Terms()
            self.terms.terms = SimpleVocabulary((
                SimpleTerm('selected', 'selected', self.label or self.field.title), ))
        return self.terms


@adapter_config(required=(IBool, IFormLayer), provides=IFieldWidget)
def SingleCheckBoxFieldWidget(field, request):  # pylint: disable=invalid-name
    """IFieldWidget factory for CheckBoxWidget."""
    widget = FieldWidget(field, SingleCheckBoxWidget(request))
    widget.label = ''  # don't show the label twice
    return widget



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.group module

This module handles groups of widgets within forms.
"""

from zope.interface import Interface, implementer

from pyams_form.events import FormUpdatedEvent
from pyams_form.form import BaseForm, get_form_weight
from pyams_form.interfaces.form import IFormContent, IGroup, IGroupForm, IGroupManager
from pyams_form.interfaces.widget import IWidgets
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config

__docformat__ = 'restructuredtext'


@implementer(IGroupManager)
class GroupManager:  # pylint: disable=no-member
    """Base groups manager mixin class"""

    groups = ()

    def get_groups(self):
        """Internal groups iterator getter"""
        registry = self.request.registry
        yield from sorted((adapter for name, adapter in
                           registry.getAdapters((self.context, self.request, self), IGroup)),
                          key=get_form_weight)

    def update(self):
        """See interfaces.IForm"""
        self.update_widgets()
        groups = []
        # static groups list
        for group_class in self.groups:
            # only instantiate the group_class if it hasn't already
            # been instantiated
            if IGroup.providedBy(group_class):
                group = group_class
            else:
                group = group_class(self.context, self.request, self)
            groups.append(group)
        # groups can also be added dynamically using adapters
        for group in self.get_groups():
            groups.append(group)
        # update all groups
        [group.update() for group in groups]  # pylint: disable=expression-not-assigned
        self.groups = tuple(groups)
        [subform.update() for subform in self.subforms]  # pylint: disable=expression-not-assigned
        [tabform.update() for tabform in self.tabforms]  # pylint: disable=expression-not-assigned
        self.request.registry.notify(FormUpdatedEvent(self))

    def extract_data(self, set_errors=True, notify=True):
        """See interfaces.IForm"""
        data, errors = super().extract_data(set_errors=set_errors, notify=notify)
        for group in self.groups:
            group_data, group_errors = group.extract_data(set_errors=set_errors, notify=notify)
            data.update(group_data)
            if group_errors:
                if errors:
                    errors += group_errors
                else:
                    errors = group_errors
        return data, errors


@implementer(IGroup)
class Group(GroupManager, BaseForm):
    """Group of field widgets within form"""

    def __init__(self, context, request, parent_form):  # pylint: disable=super-init-not-called
        self.context = context
        self.request = request
        self.parent_form = self.__parent__ = parent_form

    def update_widgets(self, prefix=None, use_form_mode=True):
        """See interfaces.IForm"""
        registry = self.request.registry
        self.widgets = registry.getMultiAdapter((self, self.request, self.get_content()),
                                                IWidgets)
        self.widgets.mode = self.mode
        if use_form_mode:
            for attr_name in ('ignore_request', 'ignore_context', 'ignore_readonly'):
                value = getattr(self.parent_form.widgets, attr_name)
                setattr(self.widgets, attr_name, value)
        if prefix is not None:
            self.widgets.prefix = prefix
        self.widgets.update()


@adapter_config(required=(Interface, IFormLayer, IGroup),
                provides=IFormContent)
def get_group_content(context, request, group):  # pylint: disable=unused-argument
    """Group content getter"""
    return group.parent_form.get_content()


@implementer(IGroupForm)
class GroupForm(GroupManager):
    """A mix-in class for add and edit forms to support groups."""

    def update(self):
        """See interfaces.IForm"""
        GroupManager.update(self)
        self.update_actions()  # pylint: disable=no-member
        self.actions.execute()  # pylint: disable=no-member



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser main module

Fanstatic resources definition.
"""

from fanstatic import Library, Resource


__docformat__ = 'restructuredtext'


# pylint: disable=invalid-name
library = Library('pyams_form', 'resources')

# pylint: disable=invalid-name
ordered_select_input = Resource(library, 'js/orderedselect-input.js',
                                minified='js/orderedselect-input.min.js')



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser.password module

Password widget implementation.
"""

from zope.interface import implementer_only
from zope.schema.interfaces import IPassword

from pyams_form.browser.text import TextWidget
from pyams_form.interfaces.widget import IPasswordWidget, IFieldWidget
from pyams_form.widget import FieldWidget
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config

__docformat__ = 'restructuredtext'


@implementer_only(IPasswordWidget)
class PasswordWidget(TextWidget):
    """Input type password widget implementation."""

    klass = 'password-widget'
    css = 'password'


@adapter_config(required=(IPassword, IFormLayer), provides=IFieldWidget)
def PasswordFieldWidget(field, request):  # pylint: disable=invalid-name
    """IFieldWidget factory for IPasswordWidget."""
    return FieldWidget(field, PasswordWidget(request))



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser.file module

This module provides default file widget.
"""

from zope.interface import implementer_only
from zope.schema.interfaces import IBytes

from pyams_form.browser.text import TextWidget
from pyams_form.interfaces.widget import IFieldWidget, IFileWidget
from pyams_form.widget import FieldWidget
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config


__docformat__ = 'restructuredtext'


@implementer_only(IFileWidget)
class FileWidget(TextWidget):
    """File input widget implementation"""

    klass = 'file-widget'
    css = 'file'

    # headers and filename attributes are set by ``IDataConverter`` to the widget
    # provided by the FileUpload object of the form
    headers = None
    filename = None

    def json_data(self):
        data = super().json_data()  # pylint: disable=bad-super-call
        data['type'] = 'file'
        return data


@adapter_config(required=(IBytes, IFormLayer),
                provides=IFieldWidget)
def FileFieldWidget(field, request):  # pylint: disable=invalid-name
    """IFieldWidget factory for FileWidget."""
    return FieldWidget(field, FileWidget(request))
